"""Encode number as base36 encoded string"""


def base_36_encode(number: int) -> str:
    """Return a base36 encoded version of the number

    :param number: a non-negative base 10 integer
    :return: base 36 string
    """
    if number < 0:
        raise ValueError("Number must be non-negative")

    if number == 0:
        return "0"

    encoded: str = ""
    while number > 0:
        number, remainder = divmod(number, 36)
        encoded = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"[remainder] + encoded

    return encoded



"""IRDI creation Plugin"""
import re
from typing import Sequence  # noqa: UP035

import rfc3987
from cmem_plugin_base.dataintegration.context import ExecutionContext
from cmem_plugin_base.dataintegration.description import Plugin, PluginParameter
from cmem_plugin_base.dataintegration.entity import (
    Entities,
    Entity,
    EntityPath,
    EntitySchema,
)
from cmem_plugin_base.dataintegration.parameter.graph import GraphParameterType
from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin
from cmem_plugin_base.dataintegration.ports import FixedNumberOfInputs, FixedSchemaPort
from cmem_plugin_base.dataintegration.utils import setup_cmempy_user_access

from cmem_plugin_irdi.components import components
from cmem_plugin_irdi.item_code import generate_item_code, init_counter

# Version Identifier
VI = "001"

PARAMETERS = [
    PluginParameter(
        name="graph",
        label="Counter graph",
        description="Graph in which the Item Code (IC) counter is stored",
        param_type=GraphParameterType(allow_only_autocompleted_values=False),
    ),
    PluginParameter(
        name="output_schema_path",
        label="Output path / property",
        description="Path or property that will connect input values and their generated IRDIs",
        default_value="http://purl.org/dc/terms/identifier",
        advanced=True,
    ),
    PluginParameter(
        name="counted_object",
        label="Counted object",
        description="The class of objects that are counted. (IRI)",
        default_value="",
        advanced=True,
    ),
    PluginParameter(
        name="input_schema_path",
        label="Input Schema Path / Property",
        description="Path from which input values are taken. "
        "If empty, values are read from the URIs of the input",
        advanced=True,
        default_value="",
    ),
] + [parameter["parameter"] for parameter in components.values()]


@Plugin(
    label="Generate base36 IRDI",
    description="Create unique [ECLASS]"
    "(https://eclass.eu/support/technical-specification/structure-and-elements/irdi) IRDIs. "
    "IRDIs are unique for each combination of (non-advanced) parameters. "
    "If no input path is configured, values are read from the URIs of the input "
    "(Transformation Input) ",
    parameters=PARAMETERS,
)
class IrdiPlugin(WorkflowPlugin):
    """IRDI Plugin"""

    def __init__(  # noqa: PLR0913
        self,
        graph: str,
        icd: str,
        oi: str,
        opi: str,
        opis: str,
        ai: str,
        csi: str,
        counted_object: str,
        input_schema_path: str,
        output_schema_path: str = "http://purl.org/dc/terms/identifier",
    ):
        self.input_schema_path = input_schema_path
        self.graph = graph
        self.output_schema_path = output_schema_path
        self.icd = icd
        self.oi = oi
        self.opi = opi.upper()
        self.opis = opis
        self.ai = ai
        self.csi = csi.upper() if csi else ""
        self.counted_object = counted_object

        # validate inputs
        for component, definition in components.items():
            value = self.__dict__.get(component)
            if value and (re.match(definition["regex"], value) is None):
                raise ValueError(component + ": wrong format")

        if self.counted_object and (rfc3987.match(self.counted_object, rule="IRI") is None):
            raise ValueError(f"Counted object: {self.counted_object} is not a valid URI")

        # define input ports if path was provided
        if self.input_schema_path:
            self.input_ports = FixedNumberOfInputs(
                [
                    FixedSchemaPort(
                        schema=EntitySchema(
                            type_uri="",
                            paths=[EntityPath(path=self.input_schema_path)],
                        )
                    )
                ]
            )

        # construct counter identifier
        self.counter = f"{self.icd}-{self.oi}-{self.opi}-{self.opis}-{self.ai}" f"#{self.csi}"

    def execute(self, inputs: Sequence[Entities], context: ExecutionContext) -> Entities | None:
        """Execute Workflow plugin"""
        setup_cmempy_user_access(context.user)

        init_counter(self.graph, self.counter, self.counted_object)

        output = []

        try:
            first_input: Entities = inputs[0]
        except IndexError as error:
            raise ValueError("Input port not connected.") from error

        schema = EntitySchema(
            type_uri=first_input.schema.type_uri, paths=[EntityPath(self.output_schema_path)]
        )

        if self.input_schema_path:
            uris = self._get_input_path(first_input, self.input_schema_path)
        else:
            uris = self._get_input_uri(first_input)

        for uri in uris:
            item_code = generate_item_code(self.graph, self.counter)
            irdi = (
                f"{self.icd}-{self.oi}-{self.opi}-{self.opis}-{self.ai}"
                f"#{self.csi}-{item_code}#{VI}"
            )
            output.append(Entity(uri=uri, values=[[irdi]]))

        return Entities(entities=output, schema=schema)

    def _get_input_uri(self, entities: Entities) -> list[str]:
        """Get URIs to be processed from URIs of input entities"""
        return [entity.uri for entity in entities.entities]

    def _get_input_path(self, entities: Entities, input_path: str) -> list[str]:
        """Get URIS to be processed from values of input entities

        :param input_path: path for which values are returned
        :raises ValueError: if input_path is not in input schema of entities
        """
        paths = entities.schema.paths
        try:
            index = next(index for index, path in enumerate(paths) if path.path == input_path)
        except StopIteration as error:
            raise ValueError(f"Input does not contain path {input_path}") from error

        return [entity.values[index][0] for entity in entities.entities]



"""Create unique item codes"""
from pathlib import Path

from cmem.cmempy.dp.proxy.graph import post
from cmem.cmempy.queries import SparqlQuery

from cmem_plugin_irdi.utils import base_36_encode

MAX_IC_LENGTH = 6
COUNTER_ONTOLOGY_GRAPH = "http://purl.org/ontology/co/core#"

INITIALIZE_COUNTER = SparqlQuery(
    text="""
    PREFIX co: <http://purl.org/ontology/co/core#>
    PREFIX dcterms: <http://purl.org/dc/terms/>
    PREFIX vs: <http://www.w3.org/2003/06/sw-vocab-status/ns#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

    WITH <{{graph}}>
    DELETE {
        ?counter co:object ?object .
    }
    INSERT {
        <{{graph}}> a <https://vocab.eccenca.com/di/Dataset> .

        ?counter a co:Counter ;
                 dcterms:identifier "{{identifier}}" ;
                 rdfs:label "{{identifier}}";
                 {{counted_object_term}}.

    } WHERE {
        BIND(SHA256("{{identifier}}") as ?identifier_encoded)
        BIND(URI(CONCAT("https://ns.eccenca.com/counter/",
                        ?identifier_encoded
                        )) as ?counter)
        OPTIONAL {
            ?counter co:object ?object .
        }
    }
    """,
    query_type="UPDATE",
)

GET_COUNT: SparqlQuery = SparqlQuery(
    text="""
    PREFIX co: <http://purl.org/ontology/co/core#>
    PREFIX dcterms: <http://purl.org/dc/terms/>
    SELECT ?count FROM <{{graph}}> WHERE {
        ?counter a co:Counter ;
                 dcterms:identifier "{{identifier}}" ;
                 co:count ?count .
    }
    """
)

UPDATE_COUNT: SparqlQuery = SparqlQuery(
    text="""
    PREFIX co: <http://purl.org/ontology/co/core#>
    PREFIX dcterms: <http://purl.org/dc/terms/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

    WITH <{{graph}}>
    DELETE {
        ?counter co:count ?count_old .
    }
    INSERT {
        ?counter co:count ?count_new .
    }
    USING <{{graph}}>
    WHERE {
        ?counter a co:Counter ;
                dcterms:identifier "{{identifier}}" .
        OPTIONAL {
            ?counter co:count ?count_old .
            BIND((?count_old + 1) as ?count_new)
        }
        OPTIONAL {
            BIND(0 as ?count_new)
        }
    }
""",
    query_type="UPDATE",
)


def generate_item_code(graph: str, identifier: str) -> str:
    """Generate a base 36 IC (item code)

    :param graph: The graph in which the counter and its value are stored
    :param identifier: A unique identifier for the counter.
    :return: A base 36 item code
    """
    placeholders = {"graph": graph, "identifier": identifier}

    UPDATE_COUNT.get_results(placeholder=placeholders)

    res = GET_COUNT.get_json_results(placeholder=placeholders)

    try:
        count = int(res["results"]["bindings"][0]["count"]["value"])
    except (KeyError, IndexError) as error:
        raise ValueError(f"No counter found for {identifier}") from error

    item_code = base_36_encode(count).zfill(MAX_IC_LENGTH)

    if len(item_code) > MAX_IC_LENGTH:
        raise ValueError(
            f"Maximum Item Code length ({MAX_IC_LENGTH}) for " f"counter {identifier} reached"
        )

    return item_code


def init_counter(graph: str, identifier: str, counted_object: str | None = None) -> None:
    """Initialize counter entity

    :param graph: The graph in which the counter is stored
    :param identifier: A unique identifier for the counter
    :param csi: Code space identifier
    """
    # Get path to vocabulary
    current_directory = Path(__file__).resolve().parent
    absolute_path = current_directory / "vocabs/counterontology.ttl"

    # Upload ontology
    post(graph=COUNTER_ONTOLOGY_GRAPH, file=absolute_path, replace=True)

    counted_object_term = f"co:object <{counted_object}>" if counted_object else ""

    INITIALIZE_COUNTER.get_results(
        placeholder={
            "graph": graph,
            "identifier": identifier,
            "counted_object_term": counted_object_term,
        }
    )



"""PluginParameters and validation regex's for IRDI components"""
from cmem_plugin_base.dataintegration.description import PluginParameter

components = {
    "icd": {
        "parameter": PluginParameter(
            name="icd", label="International Code Designator (ICD): Numeric, 4 characters"
        ),
        "regex": r"^\d{4}$",
    },
    "oi": {
        "parameter": PluginParameter(
            name="oi",
            label="Organization Identifier (OI): Numeric, 4 characters",
        ),
        "regex": r"^\d{4}$",
    },
    "opi": {
        "parameter": PluginParameter(
            name="opi",
            label="Organization Part Identifier (OPI): Alphanumeric, up to 35 "
            "characters (base36)",
            default_value="",
        ),
        "regex": r"^[a-zA-Z0-9]{0,35}$",
    },
    "opis": {
        "parameter": PluginParameter(
            name="opis",
            label="OPI Source Indicator (OPIS): Numeric, 1 character",
            default_value="",
        ),
        "regex": r"^\d$",
    },
    "ai": {
        "parameter": PluginParameter(
            name="ai",
            label="Additional information (AI): Numeric, 4 characters",
            default_value="",
        ),
        "regex": r"^\d{4}$",
    },
    "csi": {
        "parameter": PluginParameter(
            name="csi",
            label="Code-space identifier (CSI): Alphanumeric, 2 character (base36)",
        ),
        "regex": r"^[a-zA-Z0-9]{2}$",
    },
}
